/*

  Copyright <2021> <Adrian Aleksandrowicz>

  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.


  Based on the HTTPS library of Sujay Phadke ( https://github.com/electronicsguy/ESP8266/tree/master/HTTPSRedirect )

*/


#include "GCode4x8cm.h"
#include <ESP8266WiFi.h>


#include <NTPClient.h>
#include <WiFiUdp.h>
#include <EEPROM.h>


// Define NTP Client to get time
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP);

//Connection Settings
//const char* host = "script.google.com";
//const char* googleRedirHost = "script.googleusercontent.com";
//const int httpsPort = 443;

unsigned long TimeLastTimeRead;

#ifdef CREDENTIALS
const char*  ssid = mySSID;
const char* password = myPASSWORD;

#else
//Network credentials
const char*  ssid = "XXXX";
const char* password = "XXXX"; //replace with your password

#endif

// Variables to save date and time
String formattedDate;
String dayStamp;
String timeStamp;

//#define UPDATETIME 300000
#define MAX_X_TRAVEL 500 //maximale Achsenlänge [mm]
#define MAX_Y_TRAVEL 500 //[mm]
#define ResetPinEeprom 5    //EEPROM reset über Input Pin 5 : D1 gegen GND

#define CUBESIDE 80 //[mm] Seitenlänge des Quadrats/Kästchen, in das hineingeschrieben wird
#define VARIATION_DURCHSTREICHEN 15 //[mm] Variation beim Durchsteichen des Tages/Monats

#define Y_STEPS_MM_PER 80          //Aktuell eingestellt Steps/mm (Y-Achse) in Marlin (configuration.h)

long Random1Pos = 0;           //Hilfsvariable Durchstreichen-Funktion
long Random2Pos = 0;           //Hilfsvariable Durchstreichen-Funktion

int ZielXPos = 0;
int ZielYPos = 0;
                        //  1   2   3   4   5   6   7   8   9   10    11    12    13    14    15    16    17    18    19    20    21    22    23    24    25    26    27    28    29    30    31    32    33    34    35    36
int Int_StartPunktX[36] = { 0,  80, 160,240,320,400,0,  80, 160,240,  320,  400,  0,    80,   160,  240,  320,  400,  0,    80,   160,  240,  320,  400,  0,    80,   160,  240,  320,  400,  0,    80,   160,  240,  320,  400};//Startpunkt X-Achse [mm] abhängig vom Tag
int Int_StartPunktY[36] = { 400,400,400,400,400,400,320,320,320,320,  320,  320,  240,  240,  240,  240,  240,  240,  160,  160,  160,  160,  160,  160,  80,   80,   80,   80,   80,   80,   0,    0,    0,    0,    0,    0 };//Startpunkt Y-Achse [mm] abhängig vom Tag

int Monat = 0;     //INIT = 0, nur zu Testzwecken 12
int Tag = 0;       //INIT = 0
int Stunde = 0;
int Minute = 0;
int Jahr = 0;

#define MAX_EVENTS 10

long EventData[MAX_EVENTS][4] = {         //{JJJJ, MM, TT, Event)   Event: 1-> Stern (Geburtstage), 2-> Smily , 3-> Herz
                                  {2021,  01,   01,   2},       //Neujahr
                                  {2021,  12,   24,   1},       //Weihnachten
                                  {1979,  3,   4,   1},       //Geb XXX (2x Stern)
                                  {2013,  5,   6,   3},       //Geb XXX
                                  {2015,  11,   5 ,   3},       //Geb XXX
                                  {1979,  10,   21,   2},       //Geb Oma
                                  {1979,  10,   20,   1},       //Geb XXX (2x Stern)
                                  {1977,  1 ,   26,   2},       //Geb XXX
                                  {2015,  10,    3,   2},       //frei
                                  {2021,  2 ,   14,   3}};      //Valentines

int int_calenderEvent[2] = {0,0};    //2 Stellen frei
                                     //Beispiel: int_calenderEvent = {1,2} -> 1.Stelle: Arzt, 2.Stelle: Stern, int_calenderEvent = {3,0} -> 1.Stelle: Herz, 2.Stelle: Leer

byte Eeprom_Monat_Geschrieben = 0;        //Bereits geplotteter Monat
byte Eeprom_Jahr_Geschrieben = 0;         //Bereits geplottetes Jahr
byte Eeprom_Tag_Geschrieben = 0;          //Bereits geplotteter Tag
//byte Eeprom_Event_Geschrieben = 0;        //1X: Geb, 2X: Event2, 12: Geb & Event2, 10: Geb & Null

int AddressEeprom_Monat = 0;
int AddressEeprom_Jahr = 1;
int AddressEeprom_Tag = 2;


enum State_enum {ZeitDatumLesen, TagSchreiben, MonatSchreiben, JahrSchreiben,  Sleep};

State_enum state = ZeitDatumLesen;
 
//Connect to wifi
void connectToWifi() {
  Serial.print("Connecting to wifi.");
  //Serial.println(ssid);

  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
  //  Serial.print(". ");
  }
  Serial.println("");
  Serial.print("WiFiconnected");

  // Initialize a NTPClient to get time
  timeClient.begin();
  // Set offset time in seconds to adjust for your timezone, for example:
  // GMT +1 = 3600
  // GMT +8 = 28800
  // GMT -1 = -3600
  // GMT 0 = 0
  timeClient.setTimeOffset(3600);
}

void getEventData(){
int WertAktuelleStelleEvent = 0;
for (int i =0; i < MAX_EVENTS; i++) {
  if ((EventData[i][1] == Monat)&&(EventData[i][2] == Tag)){
  int_calenderEvent[WertAktuelleStelleEvent] = (EventData[i][3]);
        if (WertAktuelleStelleEvent < 2){
        WertAktuelleStelleEvent++;}
  }}
}

void ReadTimeDateFromNTP() {
  while(!timeClient.update()) {
  timeClient.forceUpdate();
}
  // The formattedDate comes with the following format:
  // 2018-05-28T16:00:13Z
  // We need to extract date and time
  formattedDate = timeClient.getFormattedDate();
  //Serial.println(formattedDate);

  // Extract date
  int splitT = formattedDate.indexOf("T");
  dayStamp = formattedDate.substring(0, splitT);
  Monat = dayStamp.substring(5, 7).toInt();
  Serial.print("Monat:");
  Serial.println(Monat);
  Tag = dayStamp.substring(8, 10).toInt();

  Serial.print("Tag:");
  Serial.println(Tag);
  Jahr = dayStamp.substring(2, 4).toInt();
  //Serial.print("Jahr: ");
  //Serial.println(Jahr);

  // Extract time
  timeStamp = formattedDate.substring(splitT+1, formattedDate.length()-1);
  //Serial.print("HOUR:");
  //Serial.println(timeStamp);
  Stunde = timeStamp.substring(0, 2).toInt();
  Serial.print("STUNDE:");
  Serial.println(Stunde);
  Minute = timeStamp.substring(5, 7).toInt();
  //Serial.print("MIN: ");
  //Serial.println(Minute);
}

void GCodeZahlSchreiben(int Zahl){
GCodeTransmit(StiftAnheben);
  switch (Zahl){
    case 0:
            GCodeTransmit(GCode_0);
  break;
    case 1:
            GCodeTransmit(GCode_1);
  break;
    case 2:
            GCodeTransmit(GCode_2);
  break;
    case 3:
            GCodeTransmit(GCode_3);
  break;
    case 4:
            GCodeTransmit(GCode_4);
  break;
     case 5:
            GCodeTransmit(GCode_5);
  break;
     case 6:
            GCodeTransmit(GCode_6);
  break;
    case 7:
            GCodeTransmit(GCode_7);
  break;
    case 8:
            GCodeTransmit(GCode_8);
  break;
  case 9:
            GCodeTransmit(GCode_9);
  break;
  }
}

void GCodeTransmit(String gcode){
int j = 0;
for (int i = 0; i < gcode.length(); i++){

 
    // Length (with one extra character for the null terminator)
    int gcode_len = gcode.length() + 1; 
 
    // Prepare the character array (the buffer) 
    char char_array[gcode_len];
 
    // Copy it over 
    gcode.toCharArray(char_array, gcode_len);
      
      if (char_array[i] != ' ')
      {
      Serial.print(char_array[i]);  
      }
      else 
      {
      Serial.println(" ");               //Neue Zeile, NLCR
      digitalWrite(LED_BUILTIN, LOW);   //LED an
      
      
      
            while(1) {
            if (Serial.available()>0){        //Warten bis daten empfangen wurden
            delay(100);
            break;
            }}
           
            String a = Serial.readString();
            if (a.indexOf("err", 0) >= 0 ){
            digitalWrite(LED_BUILTIN, LOW);   //LED an
            delay(2000);
            i = j-1;
            }
            else {
            j = i;
            digitalWrite(LED_BUILTIN, HIGH); 
            }
            
      }
}}



void GCodeSenden_PositionAnfahren(int DatumsTagAktuell, int DatumsMonatAktuell){  

switch (state)        //State_enum {ZeitDatumLesen, TagSchreiben, MonatSchreiben, JahrSchreiben,  Sleep};
    {
        case 1:       //Tag
        ZielXPos =  Int_StartPunktX[DatumsTagAktuell-1];
        ZielYPos =  Int_StartPunktY[DatumsTagAktuell-1];
        break;
        case 2:       //Monat
        ZielXPos =  Int_StartPunktX[32 + DatumsMonatAktuell%2];
        ZielYPos =  Int_StartPunktY[32 + DatumsMonatAktuell%2];
        break;
        case 3:       //Jahr
        ZielXPos =  Int_StartPunktX[34];
        ZielYPos =  Int_StartPunktY[34];
        break;
        }

  String PositionAnfahren = StiftAnheben + GCode_abs_dist + GCode_mm + FeedRateTravel + "G0X" + String(ZielXPos) + "Y" + String(ZielYPos)+ " ";
  GCodeTransmit(PositionAnfahren);
}

void GCodeSenden_TagSchreiben(){
GCodeTransmit(QuadratUmrandungRund);      //Umrandung und Punkt zur Trennung Tag/Monat mit 80mm Seitenlänge zeichnen
GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 ");     //Vor Umsakalierung in den 0-Punkt

int WertAktuelleStelleTag = 0;
int WertAktuelleStelleEvent = 0;


  for (int i = 0; i < 2; i++){      //2 Stellen
        if (int_calenderEvent[0] > 0){       //Falls min ein Event vorliegt
        GCodeTransmit("M92Y" + String((int(Y_STEPS_MM_PER/2))) + " ");    //In Y -Achse um Faktor 2 stauchen
        }
  switch (i){
        case 0:
        WertAktuelleStelleTag = Tag/10;                   //Erste Stelle des Tages schreiben
        WertAktuelleStelleEvent = int_calenderEvent[0];   //Erste Stelle des Events schreiben
        break;
        case 1:
        WertAktuelleStelleTag = Tag%10;                   //Zweite Stelle des Tages schreiben
        WertAktuelleStelleEvent = int_calenderEvent[1];   //Zweite Stelle des Events schreiben
        break;
      }
 GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 ");
 GCodeZahlSchreiben(WertAktuelleStelleTag);
 GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 ");
 GCodeTransmit("M92Y" + String((int(Y_STEPS_MM_PER))) + " ");     //Wieder Faktor 1
 switch (WertAktuelleStelleEvent){            //Kodierung der Events: 0-> Leer , 1-> Kreuz (Arzt), 2-> Stern (Geburtstage), 3-> Zahn (Zahnarzt), 4-> Smily (Restliche Einträge)
    case 0:     //Event: 1-> Stern (Geburtstage), 2-> Smily , 3-> Herz
            
  break;
    case 1:
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y40 " + ResetPos);
            GCodeTransmit(Event1_Stern);
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 G92X0Y40 ");
  break;
    case 2:
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y40 " + ResetPos);
            GCodeTransmit(Event2_Smily);
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 G92X0Y40 ");
  break;
    case 3:
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y40 " + ResetPos);
            GCodeTransmit(Event3_Herz);
            GCodeTransmit(StiftAnheben + FeedRateTravel + "G0X0Y0 G92X0Y40 ");
  break;
  }
  if (i==0){
  GCodeTransmit(StiftAnheben + NaechsteStellenPosAnfahren + ResetPos);}   //Weiter zur nächsten Stelle
  }
}


void GCodeSenden_MonatSchreiben(){
GCodeTransmit(QuadratUmrandung);      //Umrandung und Punkt zur Trennung Tag/Monat mit 80mm Seitenlänge zeichnen

int WertAktuelleStelle = 0;
  for (int i = 0; i < 2; i++){      //2 Stellen
      switch (i){
        case 0:
        WertAktuelleStelle = Monat/10;   //Erste Stelle des Tages schreiben
        break;
        case 1:
        WertAktuelleStelle = Monat%10;   //Zweite Stelle des Tages schreiben
        break;
      }
  GCodeZahlSchreiben(WertAktuelleStelle);
  if (i==0){
  GCodeTransmit(StiftAnheben + NaechsteStellenPosAnfahren + ResetPos);}   //Weiter zur nächsten Stelle
  }
}

void GCodeSenden_JahrSchreiben(){
GCodeTransmit(QuadratUmrandungBreit);      //Umrandung und Punkt zur Trennung Tag/Monat mit 80mm Seitenlänge zeichnen

int WertAktuelleStelle = 0;
  for (int i = 0; i < 4; i++){      //8 Stellen
      switch (i){
        case 0:
        WertAktuelleStelle = 2;   //Erste Stelle des Tages schreiben
        break;
        case 1:
        WertAktuelleStelle = 0;   //Zweite Stelle des Tages schreiben
        break;
        case 2:
        WertAktuelleStelle = Jahr/10;   //Erste Stelle des Tages schreiben
        break;
        case 3:
        WertAktuelleStelle = Jahr%10;   //Zweite Stelle des Tages schreiben
        break;
      }

  GCodeZahlSchreiben(WertAktuelleStelle);
  GCodeTransmit(StiftAnheben + NaechsteStellenPosAnfahren + ResetPos);   //Weiter zur nächsten Stelle
  }
}

void GCodeSenden_Homing(){
  GCodeTransmit(StiftAnheben + Homing);   //Home
}

void GCodeSenden_Durchstreichen(){        //Durchstreichen
  Random1Pos = random(0, VARIATION_DURCHSTREICHEN);           //Hilfsvariable Durchstreichen, min
  Random2Pos = random((CUBESIDE-VARIATION_DURCHSTREICHEN), CUBESIDE);           //Hilfsvariable Durchstreichen, max
  String Durchstreichen = FeedRateTravel + "G1X" + String(Random1Pos+ZielXPos)+ "Y" + String(Random1Pos+ZielYPos) + " " + "G4P100 M280P1S0 G4P100 " + FeedRateDraw + "G1X" + String(Random2Pos+ZielXPos)+ "Y" + String(Random2Pos+ZielYPos) + " " + "G1X" + String(Random1Pos+ZielXPos)+ "Y" + String(Random1Pos+ZielYPos) + " " + "G4P100 M280P1S180 G4P100 " + FeedRateTravel + "G1X" + String(Random2Pos+ZielXPos)+ "Y" + String(Random1Pos+ZielYPos) + " " + "G4P100 M280P1S0 G4P100 " + FeedRateDraw + "G1X" + String(Random1Pos+ZielXPos)+ "Y" + String(Random2Pos+ZielYPos) + " " + "G1X" + String(Random2Pos+ZielXPos)+ "Y" + String(Random1Pos+ZielYPos) + " " + "G4P100 M280P1S180 G4P100 ";
  GCodeTransmit(Durchstreichen);
}

void PrintState(){
switch (state)
{
    case 0:
  Serial.println("ZeitDatumLesen ");
  break;
    case 1:
  Serial.println("TagSchreiben ");
  break;
  case 2:
  Serial.println("MonatSchreiben ");
  break;
    case 3:
  Serial.println("JahrSchreiben ");
  break;
   case 4:
  Serial.println("Sleep ");
  break;
}}


void handle_eeprom(){
  
  if (digitalRead(ResetPinEeprom) == 0){        //EEPROM Reset
    // write a 0 to all 512 bytes of the EEPROM
  Serial.println("EEPROMArase!");
  for (int i = 0; i < 512; i++) {
    EEPROM.write(i, 0);}
  EEPROM.commit();
  Serial.println("EEPROMReset!");
  for (int i = 0; i < 6; i++) {
      digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));   //0 = LED an, LED blinkt bei reset
      delay(500);}

  }

  

}

void state_machine_run(){
  
  switch(state)
  {
    case ZeitDatumLesen:
        PrintState();
        ReadTimeDateFromNTP();           //Aktuelles Datum wird gelesen, falls Termin mit Keyword "Datum" aktiv 
        getEventData();                   
        //TimeLastTimeRead = millis();
        yield();
        state = TagSchreiben; 
        break;
 
    case TagSchreiben:
        Eeprom_Tag_Geschrieben = EEPROM.read(AddressEeprom_Tag);
        if ((Eeprom_Tag_Geschrieben == 0) || (Eeprom_Tag_Geschrieben!=Tag)){
        PrintState();
        GCodeSenden_Homing(); // Erstmal Home Position Anfahren, wenn neues Datum geschrieben wird
            if (Eeprom_Tag_Geschrieben > 0){      //Altes Datum Wegstreichen
            GCodeSenden_PositionAnfahren(Eeprom_Tag_Geschrieben, Monat);     //Alten tag anfahren
            GCodeSenden_Durchstreichen();                                   //Alten tag durchstreichen
            }
        GCodeSenden_PositionAnfahren(Tag, Monat);
        GCodeTransmit(ResetPos);
        GCodeSenden_TagSchreiben();
        state = MonatSchreiben;         //Erstmal ohne Event
        EEPROM.write(AddressEeprom_Tag, Tag);
            if (EEPROM.commit()) {
            Serial.println("EEPROM_Monat_success!");
            } 
            else {
            Serial.println("EEPROM_Monat_failed!");
            }
        }
        break;


    case MonatSchreiben:
        Eeprom_Monat_Geschrieben = EEPROM.read(AddressEeprom_Monat);
        if ((Eeprom_Monat_Geschrieben == 0) || (Eeprom_Monat_Geschrieben!=Monat)){
        PrintState();
        GCodeSenden_Homing(); // Erstmal Home Position Anfahren
            if (Eeprom_Monat_Geschrieben > 0){      //Altes Datum Wegstreichen
            GCodeSenden_PositionAnfahren(Tag, Eeprom_Monat_Geschrieben);     //Alten tag anfahren
            GCodeSenden_Durchstreichen();
            }
        GCodeSenden_PositionAnfahren(Tag, Monat);
        GCodeTransmit(ResetPos);
        GCodeSenden_MonatSchreiben();
        EEPROM.write(AddressEeprom_Monat, Monat);
            if (EEPROM.commit()) {
            Serial.println("EEPROM_Monat_success!");
            } else {
            Serial.println("EEPROM_Monat_failed!");
            }
        }
        state = JahrSchreiben;       
        
        break;

    case JahrSchreiben:
        Eeprom_Jahr_Geschrieben = EEPROM.read(AddressEeprom_Jahr);
        if ((Eeprom_Jahr_Geschrieben == 0) || (Eeprom_Jahr_Geschrieben != Jahr)){
        PrintState();
        GCodeSenden_Homing(); // Erstmal Home Position Anfahren
        GCodeSenden_PositionAnfahren(Tag, Monat);
        GCodeTransmit(ResetPos);
        GCodeSenden_JahrSchreiben();
        EEPROM.write(AddressEeprom_Jahr, Jahr);   //Letzte zwei Stellen des Jahres in EEPROM schreiben
            if (EEPROM.commit()) {
            Serial.println("EEPROM_Jahr_success!");
            } else {
            Serial.println("EEPROM_Jahr_failed!");
            }
        }
        GCodeSenden_Homing();

        state = Sleep;     
        break;

    case Sleep:
        PrintState();
        GCodeTransmit(StiftAbsenken); //Abstellen, um die Feder zu entlasten
        EEPROM.end();
        digitalWrite(LED_BUILTIN, LOW);
        delay(5000);

        ESP.deepSleep(0);
//        if (millis() > TimeLastTimeRead + UPDATETIME) {
//            state = ZeitDatumLesen;
//            
//        }
      break;
      
  }
}


void setup() {
  Serial.begin(115200);
  Serial.setTimeout(100);
  connectToWifi();
  //TimeLastTimeRead = millis();
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH);
  pinMode(ResetPinEeprom, INPUT_PULLUP);                   //Reset Eeprom, zu GND schalten
  EEPROM.begin(512);
}



void loop(){
  handle_eeprom();
  state_machine_run();
  delay(100);
}
